==> see-1.2.tar.gz: see-1.2/libsee/try.c <==
/*
 * Copyright (c) 2003
 *      David Leonard.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by David Leonard and 
 *      contributors.
 * 4. Neither the name of Mr Leonard nor the names of the contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DAVID LEONARD AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL DAVID LEONARD OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* $Id: try.c,v 1.7 2004/10/17 13:50:51 d Exp $ */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <stdio.h>
# include <stdlib.h>
#endif

#include <see/type.h>
#include <see/value.h>
#include <see/try.h>
#include <see/object.h>
#include <see/string.h>
#include <see/interpreter.h>

#include "stringdefs.h"

/*
 * Exception handling
 *
 * This implementation of exceptions uses the C stack and
 * C's setjmp/longjmp to implement exceptions. The TRY macro
 * is a sets up and tests a setjmp, and the THROW macros 
 * just do a longjmp. (That is why a garbage collector is
 * needed!)
 *
 * NOTE: I had considered using a special-return mechanism
 * for exceptions, but the complexity, performance and annoyance 
 * of the vast number of 'return checking' required and the
 * small number of expected actual exceptions during normal
 * operation made this mechanism far more favourable. Also,
 * setjmp/longjmp is part of standard POSIX and ANSI C, so
 * portability is not really lost.
 */

/**
 * Handles an uncatchable exception by aborting.
 *
 * This can happen when a SEE_THROW occurs outside of any SEE_TRY context,
 * or when an error occurs during low-level throw handling. (See 
 * SEE_error_throw(), for instance).
 */
void
SEE_throw_abort(interp, v, file, line)
	struct SEE_interpreter *interp;
	const struct SEE_value *v;
	const char *file;
	int line;
{
#ifndef NDEBUG
	fprintf(stderr, "%s:%d: threw uncatchable exception\n", file, line);
	if (v != NULL) {
	    SEE_try_context_t ctx;
	    fprintf(stderr, "  exception: ");
	    SEE_TRY(interp, ctx)
		SEE_PrintValue(interp, v, stderr);
	    if (SEE_CAUGHT(ctx))
		fprintf(stderr, "<error printing value>");
	    fprintf(stderr, "\n");
	}
#endif
	(*SEE_abort)(interp, "exception thrown but no TRY block");
}

/**
 * Converts a script 'throw location' into a string of the form "program.js:23: "
 * This string is most useful as a prefix to a message generated by the exception.
 */
struct SEE_string *
SEE_location_string(interp, loc)
	struct SEE_interpreter *interp;
	struct SEE_throw_location *loc;
{
	struct SEE_string *s;

	s = SEE_string_new(interp, 0);
	if (loc == NULL)
		return s;
	SEE_string_append(s, loc->filename ? loc->filename 
					   : STR(unknown_file));
	SEE_string_addch(s, ':');
	SEE_string_append_int(s, loc->lineno);
	SEE_string_addch(s, ':');
	SEE_string_addch(s, ' ');
	return s;
}

#ifndef NDEBUG
/**
 * A dummy function for debugging.
 * This function exists soley for debuggers to break on,
 * since the SEE_THROW macro calls longjmp directly (in order to
 * give the compiler a chance at making safer code.)
 */
void
SEE_throw()
{
}
#endif

/**
 * Handles longjmp failure by calling the interpreter abort hook.
 * This mechanism is provided by ANSI for catching problems with
 * longjmp (eg, corrupted jmp_buf).
 */
void
longjmperror()
{
	(*SEE_abort)(NULL, "longjmp error");
}

==> see-1.3.tar.gz: see-1.3/libsee/try.c <==
/*
 * Copyright (c) 2003
 *      David Leonard.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Mr Leonard nor the names of the contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DAVID LEONARD AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL DAVID LEONARD OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* $Id: try.c 898 2005-12-21 14:51:00Z d $ */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if STDC_HEADERS
# include <stdio.h>
# include <stdlib.h>
#endif

#include <see/type.h>
#include <see/value.h>
#include <see/try.h>
#include <see/object.h>
#include <see/string.h>
#include <see/debug.h>
#include <see/interpreter.h>
#include <see/system.h>

#include "stringdefs.h"
#include "dprint.h"

/*
 * Exception handling
 *
 * This implementation of exceptions uses the C stack and
 * C's setjmp/longjmp to implement exceptions. The TRY macro
 * is a sets up and tests a setjmp, and the THROW macros 
 * just do a longjmp. (That is why a garbage collector is
 * needed!)
 *
 * NOTE: I had considered using a special-return mechanism
 * for exceptions, but the complexity, performance and annoyance 
 * of the vast number of 'return checking' required and the
 * small number of expected actual exceptions during normal
 * operation made this mechanism far more favourable. Also,
 * setjmp/longjmp is part of standard POSIX and ANSI C, so
 * portability is not really lost.
 */

/**
 * Handles an uncatchable exception by aborting.
 *
 * This can happen when a SEE_THROW occurs outside of any SEE_TRY context,
 * or when an error occurs during low-level throw handling. (See 
 * SEE_error_throw(), for instance).
 */
void
SEE_throw_abort(interp, v, file, line)
	struct SEE_interpreter *interp;
	const struct SEE_value *v;
	const char *file;
	int line;
{
#ifndef NDEBUG
	dprintf("%s:%d: threw uncatchable exception\n", file, line);
	if (v != NULL) {
	    SEE_try_context_t ctx;
	    dprintf("  exception: ");
	    SEE_TRY(interp, ctx)
		dprintv(interp, v);
	    if (SEE_CAUGHT(ctx))
		dprintf("<error printing value>");
	    dprintf("\n");
	}
#endif
	SEE_ABORT(interp, "exception thrown but no TRY block");
}

/**
 * Converts a script 'throw location' into a string of the form "program.js:23: "
 * This string is most useful as a prefix to a message generated by the exception.
 */
struct SEE_string *
SEE_location_string(interp, loc)
	struct SEE_interpreter *interp;
	struct SEE_throw_location *loc;
{
	struct SEE_string *s;

	s = SEE_string_new(interp, 0);
	if (loc == NULL)
		return s;
	SEE_string_append(s, loc->filename ? loc->filename 
					   : STR(unknown_file));
	SEE_string_addch(s, ':');
	SEE_string_append_int(s, loc->lineno);
	SEE_string_addch(s, ':');
	SEE_string_addch(s, ' ');
	return s;
}

#ifndef NDEBUG
/**
 * A dummy function for debugging.
 * This function exists soley for debuggers to break on,
 * since the SEE_THROW macro calls longjmp directly (in order to
 * give the compiler a chance at making safer code.)
 */
void
SEE_throw()
{
}
#endif

/**
 * Handles longjmp failure by calling the interpreter abort hook.
 * This mechanism is provided by ANSI for catching problems with
 * longjmp (eg, corrupted jmp_buf).
 */
void
longjmperror()
{
	SEE_ABORT(NULL, "longjmp error");
}
